/*
 * Heavily based on VirtualSerial demo. Original copyright (for that demo) follows.
 *
             LUFA Library
     Copyright (C) Dean Camera, 2012.

  dean [at] fourwalledcubicle [dot] com
           www.lufa-lib.org
*/

/*
  Copyright 2012  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

#include "logger.h"

#include "onewire.h"
#include "ds18x20.h"

//#include "diskio.h"
//#include "ff.h"
#include <avr/eeprom.h> 

#define MAXSENSORS 4

uint8_t gSensorIDs[MAXSENSORS][OW_ROMCODE_SIZE];

uint16_t timer1_count = 0;
//FATFS FATFS_Obj;

// SD card functions must be called every 10 ms, which, with a prescaler of /8, is 2000 ticks.
//#define SDCARD_TIMER1_DELAY (65536 - 20000)

// Changed this to a periodic timer every 1/10th of a second (and prescaler of /64)
#define TIMER1_DELAY (65536 - 25000)
#define TICKS_PER_SEC 10
// Half a second of hysteresis
#define BUTTON_DEBOUNCE 5

#define START_EEPROM_LOCATION 4
//for testing
//#define START_EEPROM_LOCATION 1020
#define MAX_EEPROM_LOCATION 1023
#define DEFAULT_SECS_BETWEEN_READS 15

// Read from EEPROM. Stored in EEPROM as:
// uint8_t secs_between_reads
// uint8_t num_readings_high;
// uint8_t num_readings_low
// uint8_t checksum;
uint16_t ticks_between_reads = 0;
uint16_t e_write_ptr;

/** LUFA CDC Class driver interface configuration and state information. This structure is
 *  passed to all CDC Class driver functions, so that multiple instances of the same class
 *  within a device can be differentiated from one another.
 */
USB_ClassInfo_CDC_Device_t VirtualSerial_CDC_Interface =
	{
		.Config =
			{
				.ControlInterfaceNumber         = 0,

				.DataINEndpointNumber           = CDC_TX_EPNUM,
				.DataINEndpointSize             = CDC_TXRX_EPSIZE,
				.DataINEndpointDoubleBank       = false,

				.DataOUTEndpointNumber          = CDC_RX_EPNUM,
				.DataOUTEndpointSize            = CDC_TXRX_EPSIZE,
				.DataOUTEndpointDoubleBank      = false,

				.NotificationEndpointNumber     = CDC_NOTIFICATION_EPNUM,
				.NotificationEndpointSize       = CDC_NOTIFICATION_EPSIZE,
				.NotificationEndpointDoubleBank = false,
			},
	};

/** Standard file stream for the CDC interface when set up, so that the virtual CDC COM port can be
 *  used like any regular character stream in the C APIs
 */
FILE USBSerialStream;

static uint8_t search_sensors(void)
{
	uint8_t i;
	uint8_t id[OW_ROMCODE_SIZE];
	uint8_t diff, nSensors;
	
	ow_reset();

	nSensors = 0;
	
	diff = OW_SEARCH_FIRST;
	while ( diff != OW_LAST_DEVICE && nSensors < MAXSENSORS ) {
		DS18X20_find_sensor( &diff, &id[0] );
		
		if( diff == OW_PRESENCE_ERR ) {
			break;
		}
		
		if( diff == OW_DATA_ERR ) {
			break;
		}
		
		for ( i=0; i < OW_ROMCODE_SIZE; i++ )
			gSensorIDs[nSensors][i] = id[i];
		
		nSensors++;
	}
	//fprintf(&USBSerialStream, "Sensors found: %d\r\n", nSensors);
	
	return nSensors;
}

static int16_t get_decicelsius(void)
{
	int16_t decicelsius;
	uint8_t res;
	int i = gSensorIDs[0][0]; // family-code for conversion-routine
	DS18X20_start_meas( DS18X20_POWER_PARASITE, NULL );
	_delay_ms( DS18B20_TCONV_12BIT );
	res = DS18X20_read_decicelsius_single( i, &decicelsius );
	if(res != DS18X20_OK) {
		return -6000;
	}

	return decicelsius;
}

static void show_temp(void)
{
	int16_t temp;

	temp = get_decicelsius();

	if(temp == -6000) {
		fprintf(&USBSerialStream, "Read error\r\n");
	}else{
		fprintf(&USBSerialStream, "Temp: %d decicelsius\r\n", temp);
	}
}

static void show_counter(void)
{
	fprintf(&USBSerialStream, "Counter: %d\r\n", timer1_count);
}

#if 0
static void disk_test(void)
{
	FRESULT f_err_code;

	disk_initialize(0);

	fprintf(&USBSerialStream, "\r\nmount\r\n");
	f_err_code = f_mount(0, &FATFS_Obj);
	
	if(f_err_code != 0) {
		fprintf(&USBSerialStream, "\r\nmount error %d\r\n", f_err_code);
	}

	fprintf(&USBSerialStream, "\r\ncompleted\r\n");
}
#endif

static void
show_button_status(void)
{
	fprintf(&USBSerialStream, "%d\r\n", Buttons_GetStatus());
}


static void
showOptions(void)
{
	fputs("t    show temperature\r\n", &USBSerialStream);
	fputs("c    show isr counter\r\n", &USBSerialStream);
	fputs("h    show eeprom header\r\n", &USBSerialStream);
	fputs("d    show eeprom data\r\n", &USBSerialStream);
	fputs("1    data interval: 1 sec\r\n", &USBSerialStream);
	fputs("5    data interval: 16 secs\r\n", &USBSerialStream);
	fputs("6    data interval: 32 secs\r\n", &USBSerialStream);
	//fputs("\r\nd    disk test\r\n", &USBSerialStream);
}

static void
eeprom_read_header(void)
{
	uint8_t byte;
	uint8_t sum;

	byte = sum = eeprom_read_byte((uint8_t *)0);
	ticks_between_reads = byte * TICKS_PER_SEC;

	byte = eeprom_read_byte((uint8_t *)1);
	e_write_ptr = byte << 8;
	sum += byte;

	byte = eeprom_read_byte((uint8_t *)2);
	e_write_ptr |= byte;
	e_write_ptr += START_EEPROM_LOCATION;
	sum += byte;

	byte = eeprom_read_byte((uint8_t *)3);

	if(byte != sum) {
		// Reset to default values
		ticks_between_reads = DEFAULT_SECS_BETWEEN_READS * TICKS_PER_SEC;
		e_write_ptr = START_EEPROM_LOCATION;
	}
}

static void
eeprom_write_header(void)
{
	uint8_t byte;
	uint8_t sum;

	byte = sum = (ticks_between_reads / TICKS_PER_SEC);
	eeprom_write_byte((uint8_t *)0, byte);

	byte = ((e_write_ptr - START_EEPROM_LOCATION) >> 8);
	sum += byte;
	eeprom_write_byte((uint8_t *)1, byte);

	byte = ((e_write_ptr - START_EEPROM_LOCATION) & 0xff);
	sum += byte;
	eeprom_write_byte((uint8_t *)2, byte);

	eeprom_write_byte((uint8_t *)3, sum);
}

static void
eeprom_update_count(void)
{
	eeprom_write_header();
}


static void
show_header(void)
{
	fprintf(&USBSerialStream, "Secs: %d; Count: %d\r\n", ticks_between_reads / TICKS_PER_SEC, e_write_ptr - START_EEPROM_LOCATION);
}

static void
show_data(void)
{
	for(int i = START_EEPROM_LOCATION; i < e_write_ptr; i++) {
		int8_t byte = (int8_t)eeprom_read_byte((uint8_t *)i);
		fprintf(&USBSerialStream, "%d ", byte);
	}
	fputs("\r\nDone\r\n", &USBSerialStream);
}

static void
set_interval_frommenu(int byte)
{
	byte -= '0';

	ticks_between_reads = (1 << (byte - 1)) * TICKS_PER_SEC;
	
	eeprom_write_header();
	fprintf(&USBSerialStream, "secs: %d\r\n", ticks_between_reads / TICKS_PER_SEC);
}

static void
doMenu(int byte)
{
	if(byte == '\n' || byte == '?') {
		showOptions();
	} else if (byte == 't') {
		show_temp();
	} else if (byte == 'c') {
		show_counter();
	} else if (byte == 'h') {
		show_header();
	} else if (byte == 'd') {
		show_data();
	} else if (byte >= '1' && byte <= '9') {
		set_interval_frommenu(byte);
	/*} else if (byte == 'd') {
		disk_test();
	*/
	} else {
		fputs("\r\n?\r\n", &USBSerialStream);
	}
}


/** Main program entry point. This routine contains the overall program flow, including initial
 *  setup of all components and the main program loop.
 */
int main(void)
{
	SetupHardware();

	eeprom_read_header();

	/* Create a regular character stream for the interface so that it can be used with the stdio.h functions */
	CDC_Device_CreateStream(&VirtualSerial_CDC_Interface, &USBSerialStream);

	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);
	sei();

	int16_t byte;
	int16_t last_button_press = 0;
	int16_t last_recording_made = 0;
	int8_t button = 0, prevbutton = 0;
	int8_t recording_started = 0;

	search_sensors();

	for (;;)
	{
		byte = CDC_Device_ReceiveByte(&VirtualSerial_CDC_Interface);
		if(byte > 0) {
			doMenu(byte);
		}

		button = Buttons_GetStatus() & BUTTONS_BUTTON1;

		if(button && !prevbutton && (timer1_count - last_button_press > BUTTON_DEBOUNCE)) {
			last_button_press = timer1_count;
			if(recording_started) {
				// Stop recording
				recording_started = 0;
				LEDs_TurnOffLEDs(LEDS_LED1);
				eeprom_update_count();
			} else {
				// Start recording
				recording_started = 1;
				e_write_ptr = START_EEPROM_LOCATION;
			}
		}

		prevbutton = button;

		if(recording_started) {
			// Log to EEPROM
			if (timer1_count - last_recording_made >= ticks_between_reads) {
				if (e_write_ptr < MAX_EEPROM_LOCATION) {
					int16_t temp = get_decicelsius();
					int8_t byte;

					if(temp == -6000) {
						byte = 0xff;
					} else {
						byte = temp / 10;
					}
					eeprom_write_byte((uint8_t*)e_write_ptr, (uint8_t)byte);

					e_write_ptr += 1;
					last_recording_made = timer1_count;
				} else {
					recording_started = 0;
					eeprom_update_count();
					LEDs_TurnOnLEDs(LEDS_LED1);
				}
			}
		}

		if(recording_started && e_write_ptr < MAX_EEPROM_LOCATION) {
			if(timer1_count & 0x8)
				LEDs_TurnOnLEDs(LEDS_LED1);
			else
				LEDs_TurnOffLEDs(LEDS_LED1);
		}

		/* Must throw away unused bytes from the host, or it will lock up while waiting for the device */
		//CDC_Device_ReceiveByte(&VirtualSerial_CDC_Interface);

		CDC_Device_USBTask(&VirtualSerial_CDC_Interface);
		USB_USBTask();
	}
}

ISR(TIMER1_OVF_vect)
{
	TCNT1 = TIMER1_DELAY;
	timer1_count += 1;

	//disk_timerproc();
}

/* Set up the SD card timer */
void
initInterrupt(void)
{
	/* Prescaler = 8 */
	//TCCR1B |= 2;

	/* Prescaler = 64 */
	TCCR1B |= 3;

	TCNT1 = TIMER1_DELAY;
	
	/* Overflow */
    TIMSK1 |= (1 << TOIE1);
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable clock division */
	clock_prescale_set(clock_div_1);

	/* Hardware Initialization */
	LEDs_Init();
	USB_Init();
	Buttons_Init();
	initInterrupt();
}

/** Event handler for the library USB Connection event. */
void EVENT_USB_Device_Connect(void)
{
	LEDs_SetAllLEDs(LEDMASK_USB_ENUMERATING);
}

/** Event handler for the library USB Disconnection event. */
void EVENT_USB_Device_Disconnect(void)
{
	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);
}

/** Event handler for the library USB Configuration Changed event. */
void EVENT_USB_Device_ConfigurationChanged(void)
{
	bool ConfigSuccess = true;

	ConfigSuccess &= CDC_Device_ConfigureEndpoints(&VirtualSerial_CDC_Interface);

	LEDs_SetAllLEDs(ConfigSuccess ? LEDMASK_USB_READY : LEDMASK_USB_ERROR);
}

/** Event handler for the library USB Control Request reception event. */
void EVENT_USB_Device_ControlRequest(void)
{
	CDC_Device_ProcessControlRequest(&VirtualSerial_CDC_Interface);
}

